#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import grp
import re
import pwd
import subprocess
import logging

from ..command import Command
from ..tools import question, get_current_user, mkdir


class InitCommand(Command):

    _arguments = [
        (
            ['-f', '--force'],
            {
                'action': 'store_true',
                'help': 'Force the initialisation'
            }
        )
    ]
    _description = 'Initialise MDK'

    def resolve_directory(self, path, user):
        if path.startswith('~'):
            path = re.sub(r'^~', '~%s' % user, path)
        path = os.path.abspath(os.path.realpath(os.path.expanduser(path)))
        return path

    def run(self, args):

        # Check what user we want to initialise for.
        while True:
            username = question('What user are you initialising MDK for?', get_current_user())
            try:
                user = pwd.getpwnam(username)
            except:
                logging.warning('Error while getting information for user %s' % (username))
                continue

            try:
                usergroup = grp.getgrgid(user.pw_gid)
            except:
                logging.warning('Error while getting the group of user %s' % (username))
                continue

            break

        # Default directories.
        userdir = self.resolve_directory('~/.moodle-sdk', username)
        scriptdir = os.path.dirname(os.path.realpath(__file__))

        # Create the main MDK folder.
        if not os.path.isdir(userdir):
            logging.info('Creating directory %s.' % userdir)
            mkdir(userdir, 0755)
            os.chown(userdir, user.pw_uid, usergroup.gr_gid)

        # Checking if the config file exists.
        userconfigfile = os.path.join(userdir, 'config.json')
        if os.path.isfile(userconfigfile):
            logging.info('Config file %s already in place.' % userconfigfile)
            if not args.force:
                raise Exception('Aborting. Use --force to continue.')

        elif not os.path.isfile(userconfigfile):
            logging.info('Creating user config file in %s.' % userconfigfile)
            open(userconfigfile, 'w')
            os.chown(userconfigfile, user.pw_uid, usergroup.gr_gid)

        # Loading the configuration.
        from ..config import Conf as Config
        C = Config(userfile=userconfigfile)

        # Asks the user what needs to be asked.
        while True:
            www = question('What is the DocumentRoot of your virtual host?', C.get('dirs.www'))
            www = self.resolve_directory(www, username)
            try:
                if not os.path.isdir(www):
                    mkdir(www, 0775)
                    os.chown(www, user.pw_uid, usergroup.gr_gid)
            except:
                logging.error('Error while creating directory %s' % www)
                continue

            if not os.access(www, os.W_OK):
                logging.error('You need to have permission to write to that directory.\nPlease fix or use another directory.')
                continue

            C.set('dirs.www', www)
            break

        while True:
            storage = question('Where do you want to store your Moodle instances?', C.get('dirs.storage'))
            storage = self.resolve_directory(storage, username)
            try:
                if not os.path.isdir(storage):
                    if storage != www:
                        mkdir(storage, 0775)
                        os.chown(storage, user.pw_uid, usergroup.gr_gid)
                    else:
                        logging.error('Error! dirs.www and dirs.storage must be different!')
                        continue
            except:
                logging.error('Error while creating directory %s' % storage)
                continue

            if not os.access(storage, os.W_OK):
                logging.error('You need to have permission to write to that directory.\nPlease fix or use another directory.')
                continue

            C.set('dirs.storage', storage)
            break

        # The default configuration file should point to the right directory for dirs.mdk,
        # we will just ensure that it exists.
        mdkdir = C.get('dirs.mdk')
        mdkdir = self.resolve_directory(mdkdir, username)
        if not os.path.isdir(mdkdir):
            try:
                logging.info('Creating MDK directory %s' % mdkdir)
                mkdir(mdkdir, 0775)
                os.chown(mdkdir, user.pw_uid, usergroup.gr_gid)
            except:
                logging.error('Error while creating %s, please fix manually.' % mdkdir)

        # Git repository.
        github = question('What is your Github username? (Leave blank if not using Github)')
        if github != None:
            C.set('remotes.mine', C.get('remotes.mine').replace('YourGitHub', github))
            C.set('repositoryUrl', C.get('repositoryUrl').replace('YourGitHub', github))
            C.set('diffUrlTemplate', C.get('diffUrlTemplate').replace('YourGitHub', github))
            C.set('myRemote', 'github')
            C.set('upstreamRemote', 'origin')
        else:
            C.set('remotes.mine', question('What is your remote?', C.get('remotes.mine')))
            C.set('myRemote', question('What to call your remote?', C.get('myRemote')))
            C.set('upstreamRemote', question('What to call the upsream remote (official Moodle remote)?', C.get('upstreamRemote')))

        # Database settings.
        C.set('db.mysqli.user', question('What is your MySQL user?', C.get('db.mysqli.user')))
        C.set('db.mysqli.passwd', question('What is your MySQL password?', 'root', password=True))
        C.set('db.pgsql.user', question('What is your PostgreSQL user?', C.get('db.pgsql.user')))
        C.set('db.pgsql.passwd', question('What is your PostgreSQL password?', 'root', password=True))

        print ''
        print 'MDK has been initialised with minimal configuration.'
        print 'For more settings, edit your config file: %s.' % userconfigfile
        print 'Use %s as documentation.' % os.path.join(scriptdir, 'config-dist.json')
        print ''
        print 'Type the following command to create your first instance:'
        print '  mdk create'
        print '(This will take some time, but don\'t worry, that\'s because the cache is still empty)'
        print ''
        print '/!\ Please logout/login before to avoid permission issues: sudo su `whoami`'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2014 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
import os
from .config import Conf
from .tools import mkdir, process

C = Conf()


class PHPUnit(object):
    """Class wrapping PHPUnit functions"""

    _M = None
    _Wp = None

    def __init__(self, Wp, M):
        self._Wp = Wp
        self._M = M

    def getCommand(self, testcase=None, unittest=None, filter=None, coverage=None, testsuite=None, stopon=None):
        """Get the PHPUnit command"""
        cmd = []
        if self.usesComposer():
            cmd.append('vendor/bin/phpunit')
        else:
            cmd.append('phpunit')

        if coverage:
            cmd.append('--coverage-html')
            cmd.append(self.getCoverageDir())

        if stopon:
            for on in stopon:
                cmd.append('--stop-on-%s' % on)

        if testcase:
            cmd.append(testcase)
        elif unittest:
            cmd.append(unittest)
        elif filter:
            cmd.append('--filter="%s"' % filter)
        elif testsuite:
            cmd.append('--testsuite')
            cmd.append(testsuite)

        return cmd

    def getCoverageDir(self):
        """Get the Coverage directory, and create it if required"""
        return self.Wp.getExtraDir(self.M.get('identifier'), 'coverage')

    def getCoverageUrl(self):
        """Return the code coverage URL"""
        return self.Wp.getUrl(self.M.get('identifier'), extra='coverage')

    def init(self, force=False, prefix=None):
        """Initialise the PHPUnit environment"""

        if self.M.branch_compare(23, '<'):
            raise Exception('PHPUnit is only available from Moodle 2.3')

        # Set PHPUnit data root
        phpunit_dataroot = self.M.get('dataroot') + '_phpu'
        self.M.updateConfig('phpunit_dataroot', phpunit_dataroot)
        if not os.path.isdir(phpunit_dataroot):
            mkdir(phpunit_dataroot, 0777)

        # Set PHPUnit prefix
        currentPrefix = self.M.get('phpunit_prefix')
        phpunit_prefix = prefix or 'phpu_'

        if not currentPrefix or force:
            self.M.updateConfig('phpunit_prefix', phpunit_prefix)
        elif currentPrefix != phpunit_prefix and self.M.get('dbtype') != 'oci':
            # Warn that a prefix is already set and we did not change it.
            # No warning for Oracle as we need to set it to something else.
            logging.warning('PHPUnit prefix not changed, already set to \'%s\', expected \'%s\'.' % (currentPrefix, phpunit_prefix))

        result = (None, None, None)
        exception = None
        try:
            if force:
                result = self.M.cli('/admin/tool/phpunit/cli/util.php', args='--drop', stdout=None, stderr=None)
            result = self.M.cli('/admin/tool/phpunit/cli/init.php', stdout=None, stderr=None)
        except Exception as exception:
            pass

        if exception != None or result[0] > 0:
            if result[0] == 129:
                raise Exception('PHPUnit is not installed on your system')
            elif result[0] > 0:
                raise Exception('Something wrong with PHPUnit configuration')
            else:
                raise exception

        logging.info('PHPUnit ready!')

    def run(self, **kwargs):
        """Execute the command"""
        cmd = self.getCommand(**kwargs)
        return process(cmd, self.M.get('path'), None, None)

    def usesComposer(self):
        """Return whether or not the instance uses composer, the latter is considered installed"""
        return os.path.isfile(os.path.join(self.M.get('path'), 'composer.json'))

    @property
    def M(self):
        return self._M

    @property
    def Wp(self):
        return self._Wp



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2012 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import json
import re
import copy


class ConfigObject(object):
    """Configuration object"""
    data = None

    def __init__(self):
        self.data = {}

    def add(self, name, value):
        """Add a new config but throws an exception if already defined"""
        if self.get(name) != None:
            raise Exception('Setting already declared')
        self.set(name, value)

    def get(self, name=None, default=None):
        """Return all the settings, or the setting if name is specified.
        In case the setting is not found default is returned instead.
        """
        data = copy.deepcopy(self.data)
        if name != None:
            name = unicode(name).split('.')
            for n in name:
                try:
                    data = data[n]
                except:
                    data = default
                    break
        return data

    def getFlat(self, data=None, parent=''):
        """Return the entire data as a flat array"""
        flatten = {}
        if data == None:
            data = self.get()
        for k, v in data.items():
            newKey = '%s.%s' % (parent, k) if parent != '' else k
            if type(v) == dict:
                for k2, v2 in self.getFlat(v, newKey).items():
                    flatten[k2] = v2
            else:
                flatten[newKey] = v
        return flatten

    def load(self, data, merge=False):
        """Load up the data"""
        if merge:
            data = self.mergeData(self.data, data)
        self.data = data

    def loadFromFile(self, filepath, merge=False):
        """Load the settings from a file"""
        if not os.path.isfile(filepath):
            raise ConfigFileNotFound('Could not find the config file %s' % filepath)
        try:
            lines = ''
            f = open(filepath, 'r')
            for l in f:
                if re.match(r'^\s*//', l):
                    continue
                lines += l
            if len(lines) > 0:
                self.load(json.loads(lines), merge=merge)
            f.close()
        except:
            raise ConfigFileCouldNotBeLoaded('Could not load config file %s' % filepath)

    def mergeData(self, origData, newData):
        """Recursively merge 2 dict of data"""
        for k, v in newData.items():
            if k in origData and type(v) == dict:
                origData[k] = self.mergeData(origData[k], v)
            else:
                # Deepcopy prevents references to original object
                origData[k] = copy.deepcopy(v)
        return origData

    def remove(self, name):
        """Remove a setting"""
        name = unicode(name).split('.')
        count = len(name)
        data = self.data
        for i in range(count):
            n = name[i]
            if i == (count - 1):
                try:
                    del data[n]
                except:
                    pass
                break
            else:
                try:
                    data = data[n]
                except:
                    break

    def set(self, name, value):
        """Set a new setting"""
        if type(value) == str:
            value = unicode(value)
        name = unicode(name).split('.')
        count = len(name)
        data = self.data
        for i in range(count):
            n = name[i]
            if i == (count - 1):
                data[n] = value
                break
            else:
                try:
                    data = data[n]
                except:
                    data[n] = {}
                    data = data[n]


class Config(object):
    """Generic config class"""

    files = None
    _loaded = False

    # ConfigObject storing a merge of all the config files
    data = None

    # ConfigObject for each config file
    objects = None

    def __init__(self, files=[]):
        """Creates the configuration object"""
        self.files = []
        for f in files:
            self.files.append(f)
        self.data = ConfigObject()
        self.objects = {}

    def add(self, name, value):
        """Add a new config"""
        self.data.add(name, value)

    def get(self, name=None):
        """Return a setting"""
        return self.data.get(name)

    def load(self, allowMissing=False):
        """Loads the configuration from the config files"""

        if self._loaded:
            return True

        for fn in self.files:
            self.objects[fn] = ConfigObject()
            try:
                self.objects[fn].loadFromFile(fn)
            except ConfigFileNotFound as e:
                if not allowMissing:
                    raise e
            self.data.load(self.objects[fn].get(), merge=True)

    def reload(self):
        """Reload the configuration"""
        self._loaded = False
        self.load()

    def remove(self, name):
        """Remove a setting"""
        self.data.remove(name)

    def save(self, to, confObj=None):
        """Save the settings to the config file"""
        if not confObj:
            confObj = self.data
        try:
            f = open(to, 'w')
            json.dump(confObj.get(), f, indent=4)
            f.close()
        except Exception as e:
            print e
            raise ConfigFileCouldNotBeSaved('Could not save to config file %s' % to)

    def set(self, name, value):
        """Set a new setting"""
        self.data.set(name, value)


class Conf(Config):
    """MDK config class"""

    userFile = None
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Conf, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self, userfile=None):
        if userfile == None:
            userfile = os.path.expanduser('~/.moodle-sdk/config.json')
        self.userFile = userfile
        files = [
            os.path.join(os.path.dirname(__file__), 'config-dist.json'),
            os.path.join(os.path.dirname(__file__), 'config.json'),
            '/etc/moodle-sdk/config.json',
            self.userFile,
        ]
        Config.__init__(self, files)
        self.load(allowMissing=True)

    def add(self, name, value):
        """Set a new setting"""
        super(Conf, self).add(name, value)
        self.save()

    def remove(self, name):
        """Remove a setting"""
        # We need to remove from the object we use as a reference in save().
        self.objects[self.userFile].remove(name)
        super(Conf, self).remove(name)
        self.save()

    def save(self, to=None, confObj=None):
        """Save only the difference to the user config file"""

        # The base file to use is the user file
        to = self.userFile
        diffData = self.objects[self.userFile]

        files = list(self.files)
        files.reverse()

        # Each of the know settings will be checked
        data = self.data.getFlat()
        for k in sorted(data.keys()):
            v = data[k]
            different = False
            found = False

            # Respect the files order when reading the settings
            for f in files:
                o = self.objects[f]
                ov = o.get(k)

                # The value hasn't been found and is different
                if not found and ov != None and ov != v:
                    different = True
                    break
                # The value is set
                elif ov != None and ov == v:
                    found = True

            # The value differs, or none of the file define it
            if different or not found:
                diffData.set(k, v)

        confObj = diffData
        super(Conf, self).save(to, confObj)

    def set(self, name, value):
        """Set a new setting"""
        super(Conf, self).set(name, value)
        self.save()


class ConfigFileNotFound(Exception):
    pass


class ConfigFileCouldNotBeLoaded(Exception):
    pass


class ConfigFileCouldNotBeSaved(Exception):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2012 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
import MySQLdb as mysql
import psycopg2 as pgsql


# TODO: Clean up the mess caused by the different engines and libraries.


class DB(object):

    conn = None
    cur = None
    engine = None
    options = None

    def __init__(self, engine, options):

        self.engine = engine
        self.options = options

        if engine in ('mysqli', 'mariadb'):

            if 'fuckfred' in options['passwd']:
                raise Exception('Could not establish connexion with MySQL, bad language used!')

            self.conn = mysql.connect(
                host=options['host'],
                port=int(options['port']),
                user=options['user'],
                passwd=options['passwd'],
                db=''
            )
            self.cur = self.conn.cursor()

        elif engine == 'pgsql':
            # psycopg2.
            self.conn = pgsql.connect(
                host=str(options['host']),
                port=int(options['port']),
                user=str(options['user']),
                password=str(options['passwd'])
            )
            try:
                self.cur = self.conn.cursor()
            except:
                raise Exception('Connexion failed! Make sure the database \'%s\' exists.' % str(options['user']))

        else:
            raise Exception('DB engine %s not supported' % engine)

    def close(self):
        """Close the cursor and connection"""
        self.cur.close()
        self.conn.close()

    def columns(self, table):

        columns = []

        if self.engine in ('mysqli', 'mariadb'):
            logging.debug('DESCRIBE %s' % table)
            self.cur.execute('DESCRIBE %s' % table)
            for column in self.cur.fetchall():
                columns.append(column[0])

        return columns

    def createdb(self, db):

        try:
            # Disable transaction on PostgreSQL.
            old_isolation_level = self.conn.isolation_level
            self.conn.set_isolation_level(0)
        except:
            pass

        if self.engine in ('mysqli', 'mariadb'):
            sql = 'CREATE DATABASE `%s` DEFAULT CHARACTER SET utf8 COLLATE utf8_unicode_ci' % db
        elif self.engine == 'pgsql':
            sql = 'CREATE DATABASE "%s" WITH ENCODING \'UNICODE\'' % db

        logging.debug(sql)
        self.cur.execute(sql)

        try:
            self.conn.set_isolation_level(old_isolation_level)
        except:
            pass

    def dbexists(self, db):

        count = None
        if self.engine in ('mysqli', 'mariadb'):
            sql = "SELECT COUNT('*') FROM information_schema.SCHEMATA WHERE SCHEMA_NAME LIKE '%s'" % db

        elif self.engine == 'pgsql':
            sql = "SELECT COUNT('*') FROM pg_database WHERE datname='%s'" % db

        logging.debug(sql)
        self.cur.execute(sql)
        count = self.cur.fetchone()[0]

        return count > 0

    def dropdb(self, db):

        try:
            # Disable transaction on PostgreSQL.
            old_isolation_level = self.conn.isolation_level
            self.conn.set_isolation_level(0)
        except:
            pass

        if self.engine in ('mysqli', 'mariadb'):
            sql = 'DROP DATABASE `%s`' % db
        elif self.engine == 'pgsql':
            sql = 'DROP DATABASE "%s"' % db

        logging.debug(sql)
        self.cur.execute(sql)

        try:
            self.conn.set_isolation_level(old_isolation_level)
        except:
            pass

    def dump(self, fd, prefix=''):
        """Dump a database to the file descriptor passed"""

        if self.engine not in ('mysqli', 'mariadb'):
            raise Exception('Function dump not supported by %s' % self.engine)
        if not type(fd) == file:
            raise Exception('Passed parameter is not a file object')

        # Looping over selected tables
        tables = self.tables()
        for table in tables:
            if prefix != '' and not table.startswith(prefix):
                continue

            self.cur.execute('SHOW CREATE TABLE %s' % table)
            schema = self.cur.fetchone()[1]
            fd.write(schema + ';\n')

            # Get the columns
            columns = self.columns(table)

            # Get the field values
            self.cur.execute('SELECT %s FROM %s' % (','.join(columns), table))
            for row in self.cur.fetchall():
                values = []
                for value in row:
                    if value == None:
                        value = 'null'
                    else:
                        value = str(self.conn.escape(value))
                    values.append(value)
                insert = 'INSERT INTO %s (%s) VALUES(%s)' % (table, ','.join(columns), ','.join(values))
                fd.write(insert + ';\n')

    def execute(self, query):
        self.cur.execute(query)
        # TODO: force transaction to be executed?

    def selectdb(self, db):

        if self.engine in ('mysqli', 'mariadb'):
            self.cur.execute('USE %s' % db)

        elif self.engine == 'pgsql':
            if self.cur:
                self.cur.close()
            if self.conn:
                self.conn.close()

            # psycopg2.
            self.conn = pgsql.connect(
                host=str(self.options['host']),
                port=int(self.options['port']),
                user=str(self.options['user']),
                password=str(self.options['passwd']),
                database=str(db)
            )


            self.cur = self.conn.cursor()

    def tables(self):

        tables = []

        if self.engine in ('mysqli', 'mariadb'):
            self.cur.execute('SHOW TABLES')
            for row in self.cur.fetchall():
                tables.append(row[0])

        return tables



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2014 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import sys
import logging
from .. import tools, jira
from ..ci import CI, CIException
from ..command import Command


class PrecheckCommand(Command):

    _arguments = [
        (
            ['-b', '--branch'],
            {
                'metavar': 'branch',
                'help': 'the branch to pre-check. Defaults to the current branch.'
            }
        ),
        (
            ['-p', '--push'],
            {
                'action': 'store_true',
                'help': 'if set, the branch will be pushed to your default remote.'
            }
        ),
        (
            ['name'],
            {
                'default': None,
                'help': 'name of the instance',
                'metavar': 'name',
                'nargs': '?'
            }
        )
    ]
    _description = 'Pre-checks a branch on the CI server'

    FAILED = -1

    def run(self, args):
        M = self.Wp.resolve(args.name)
        if not M:
            raise Exception('This is not a Moodle instance')

        against = M.get('stablebranch')
        branch = args.branch or M.currentBranch()
        if branch == 'HEAD':
            raise Exception('Cannot pre-check the HEAD branch')
        elif branch == against:
            raise Exception('Cannot pre-check the stable branch')

        parsedbranch = tools.parseBranch(branch)
        if not parsedbranch:
            raise Exception('Could not parse the branch')

        issue = parsedbranch['issue']

        if args.push:
            J = jira.Jira()
            if J.isSecurityIssue('MDL-%s' % (issue)):
                raise Exception('Security issues cannot be pre-checked')

            remote = self.C.get('myRemote')
            logging.info('Pushing branch \'%s\' to remote \'%s\'', branch, remote)
            result = M.git().push(remote, branch)
            if result[0] != 0:
                raise Exception('Could not push the branch:\n  %s' % result[2])

        ci = CI()
        try:
            # TODO Remove that ugly hack to get the read-only remote.
            logging.info('Invoking the build on the CI server...')
            (outcome, infos) = ci.precheckRemoteBranch(self.C.get('repositoryUrl'), branch, against, 'MDL-%s' % issue)
        except CIException as e:
            raise e


        if outcome == CI.FAILURE:
            logging.warning('Build failed, please refer to:\n  %s', infos.get('url', '[Unknown URL]'))
            sys.exit(self.FAILED)

        elif outcome == CI.SUCCESS:
            logging.info('Precheck passed, good work!')
            sys.exit(0)


        # If we get here, that was a fail.
        if outcome == CI.ERROR:
            logging.info('Precheck FAILED with ERRORS.')
        else:
            logging.info('Precheck FAILED with WARNINGS.')
        logging.info('')

        mapping = {
            'phplint': 'PHP Lint',
            'phpcs': 'PHP coding style',
            'js': 'Javascript',
            'css': 'CSS',
            'phpdoc': 'PHP Doc',
            'commit': 'Commit message',
            'savepoint': 'Update/Upgrade',
            'thirdparty': 'Third party',
            'grunt': 'Grunt',
            'shifter': 'Shifter',
            'travis': 'Travis CI'
        }

        for key in mapping:
            try:
                details = infos.get(key)
            except KeyError:
                logging.debug("Unknown key '%s', ignoring...", key)
                continue

            result = details.get('result', None)
            symbol = ' ' if result == CI.SUCCESS else ('!' if result == CI.WARNING else 'X')
            print '  [{}] {:<20}({} errors, {} warnings)'.format(symbol, mapping.get(key, key), details.get('errors', '?'), details.get('warnings', '?'))

        logging.info('')
        logging.info('More details at: %s', infos.get('url'))
        sys.exit(self.FAILED)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import argparse
import sys


class Command(object):
    """Represents a command"""

    _arguments = [
            (
                ['foo'],
                {
                    'help': 'I\'m an argument'
                }
            ),
            (
                ['-b', '--bar'],
                {
                    'action': 'store_true',
                    'help': 'I\'m a flag'
                }
            )
        ]
    _description = 'Undocumented command'

    __C = None
    __Wp = None

    def __init__(self, config):
        self.__C = config

    def argumentError(self, message):
        raise CommandArgumentError(message)

    @property
    def arguments(self):
        return self._arguments

    @property
    def C(self):
        return self.__C

    @property
    def description(self):
        return self._description

    def run(self, args):
        return True

    @property
    def Wp(self):
        if not self.__Wp:
            from .workplace import Workplace
            self.__Wp = Workplace()
        return self.__Wp


class CommandArgumentError(Exception):
    """Exception when a command sends an argument error"""
    pass


class CommandArgumentFormatter(argparse.HelpFormatter):
    """Custom argument formatter"""

    def _get_help_string(self, action):
        help = action.help
        if '%(default)' not in action.help:
            forbiddentypes = ['_StoreTrueAction', '_StoreFalseAction']
            if action.__class__.__name__ not in forbiddentypes and action.default is not argparse.SUPPRESS:
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    help += ' (default: %(default)s)'
        return help


class CommandArgumentParser(argparse.ArgumentParser):
    """Custom argument parser"""

    def error(self, message):
        self.print_help(sys.stderr)
        self.exit(2, '\n%s: error: %s\n' % (self.prog, message))


class CommandRunner(object):
    """Executes a command"""

    def __init__(self, command):
        self._command = command

    @property
    def command(self):
        return self._command

    def run(self, sysargs=sys.argv, prog=None):
        parser = CommandArgumentParser(description=self.command.description, prog=prog,
            formatter_class=CommandArgumentFormatter)
        for argument in self.command.arguments:
            args = argument[0]
            kwargs = argument[1]
            if 'sub-commands' in kwargs:
                subs = kwargs['sub-commands']
                del kwargs['sub-commands']
                subparsers = parser.add_subparsers(**kwargs)
                for name, sub in subs.items():
                    subparser = subparsers.add_parser(name, **sub[0])
                    defaults = {args[0]: name}
                    subparser.set_defaults(**defaults)
                    for subargument in sub[1]:
                        sargs = subargument[0]
                        skwargs = subargument[1]
                        if skwargs.has_key('silent'):
                            del skwargs['silent']
                            skwargs['help'] = argparse.SUPPRESS
                        subparser.add_argument(*sargs, **skwargs)
            else:
                if kwargs.has_key('silent'):
                    del kwargs['silent']
                    kwargs['help'] = argparse.SUPPRESS
                parser.add_argument(*args, **kwargs)
        args = parser.parse_args(sysargs)

        try:
            self.command.run(args)
        except CommandArgumentError as e:
            parser.error(e.message)


if __name__ == "__main__":
    CommandRunner(Command()).run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2014 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
from jenkinsapi import jenkins
from jenkinsapi.custom_exceptions import JenkinsAPIException, TimeOut
from .config import Conf

C = Conf()


class CI(object):
    """Wrapper for Jenkins"""

    SUCCESS = 'S'
    FAILURE = 'F'
    ERROR = 'E'
    WARNING = 'W'

    _jenkins = None
    url = None
    token = None

    def __init__(self, url=None, token=None, load=True):
        self.url = url or C.get('ci.url')
        self.token = token or C.get('ci.token')
        if load:
            self.load()

    @property
    def jenkins(self):
        """The Jenkins object"""
        return self._jenkins

    def load(self):
        """Loads the Jenkins object"""

        # Resets the logging level.
        logger = logging.getLogger('jenkinsapi.job')
        logger.setLevel(logging.WARNING)
        logger = logging.getLogger('jenkinsapi.build')
        logger.setLevel(logging.WARNING)

        # Loads the jenkins object.
        self._jenkins = jenkins.Jenkins(self.url)

    def precheckRemoteBranch(self, remote, branch, integrateto, issue=None):
        """Runs the precheck job and returns the outcome"""
        params = {
            'remote': remote,
            'branch': branch,
            'integrateto': integrateto
        }
        if issue:
            params['issue'] = issue

        job = self.jenkins.get_job('Precheck remote branch')

        try:
            invoke = job.invoke(build_params=params, securitytoken=self.token, invoke_pre_check_delay=0)
            invoke.block_until_not_queued(60, 2)
        except TimeOut:
            raise CIException('The build has been in queue for more than 60s. Aborting, please refer to: %s' % job.baseurl)
        except JenkinsAPIException:
            raise CIException('Failed to invoke the build, check your permissions.')

        build = invoke.get_build()

        logging.info('Waiting for the build to complete, please wait...')
        build.block_until_complete(3)

        # Checking the build
        outcome = CI.SUCCESS
        infos = {'url': build.baseurl}

        if build.is_good():
            logging.debug('Build complete, checking precheck results...')

            output = build.get_console()
            result = self.parseSmurfResult(output)
            if not result:
                outcome = CI.FAILURE
            else:
                outcome = result['smurf']['result']
                infos = dict(infos.items() + result.items())

        else:
            outcome = CI.FAILURE

        return (outcome, infos)

    def parseSmurfResult(self, output):
        """Parse the smurt result"""
        result = {}

        for line in output.splitlines():
            if not line.startswith('SMURFRESULT'):
                continue

            line = line.replace('SMURFRESULT: ', '')
            (smurf, rest) = line.split(':')
            elements = [smurf]
            elements.extend(rest.split(';'))
            for element in elements:
                data = element.split(',')

                errors = int(data[2])
                warnings = int(data[3])

                if errors > 0:
                    outcome = CI.ERROR
                elif warnings > 0:
                    outcome = CI.WARNING
                else:
                    outcome = CI.SUCCESS

                result[data[0]] = {
                    'errors': errors,
                    'warnings': warnings,
                    'result': outcome
                }

            break

        return result


class CIException(Exception):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import logging
from .. import db
from ..command import Command
from ..tools import mkdir

DB = db.DB


class InstallCommand(Command):

    _description = 'Install a Moodle instance'

    def __init__(self, *args, **kwargs):
        super(InstallCommand, self).__init__(*args, **kwargs)
        self._arguments = [
            (
                ['-e', '--engine'],
                {
                'action': 'store',
                  'choices': ['mariadb', 'mysqli', 'pgsql'],
                  'default': self.C.get('defaultEngine'),
                  'help': 'database engine to use',
                  'metavar': 'engine'
                }
            ),
            (
                ['-f', '--fullname'],
                {
                    'action': 'store',
                    'help': 'full name of the instance',
                    'metavar': 'fullname'
                }
            ),
            (
                ['-r', '--run'],
                {
                    'action': 'store',
                    'help': 'scripts to run after installation',
                    'metavar': 'run',
                    'nargs': '*'
                }
            ),
            (
                ['name'],
                {
                    'default': None,
                    'help': 'name of the instance',
                    'metavar': 'name',
                    'nargs': '?'
                })
        ]

    def run(self, args):

        name = args.name
        engine = args.engine
        fullname = args.fullname

        M = self.Wp.resolve(name)
        if not M:
            raise Exception('This is not a Moodle instance')

        name = M.get('identifier')
        dataDir = self.Wp.getPath(name, 'data')
        if not os.path.isdir(dataDir):
            mkdir(dataDir, 0777)

        kwargs = {
            'engine': engine,
            'fullname': fullname,
            'dataDir': dataDir,
            'wwwroot': self.Wp.getUrl(name)
        }
        M.install(**kwargs)

        # Running scripts
        if M.isInstalled() and type(args.run) == list:
            for script in args.run:
                logging.info('Running script \'%s\'' % (script))
                try:
                    M.runScript(script)
                except Exception as e:
                    logging.warning('Error while running the script: %s' % e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2012 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import sys
import os
import signal
import subprocess
import shlex
import re
import threading
import getpass
import logging
import hashlib
import tempfile
from .config import Conf

C = Conf()

def yesOrNo(q):
    while True:
        i = raw_input('%s (y/n) ' % (q)).strip().lower()
        if i == 'y':
            return True
        elif i == 'n':
            return False


def question(q, default=None, options=None, password=False):
    """Asks the user a question, and return the answer"""
    text = q
    if default != None:
        text = text + ' [%s]' % str(default)

    if password:
        i = getpass.getpass('%s\n   ' % text)
    else:
        i = raw_input('%s\n  ' % text)

    if i.strip() == '':
        return default
    else:
        if options != None and i not in options:
            return question(q, default, options)
        return i


def chmodRecursive(path, chmod):
    os.chmod(path, chmod)
    for (dirpath, dirnames, filenames) in os.walk(path):
        for d in dirnames:
            dir = os.path.join(dirpath, d)
            os.chmod(dir, chmod)
        for f in filenames:
            file = os.path.join(dirpath, f)
            os.chmod(file, chmod)


def getMDLFromCommitMessage(message):
    """Return the MDL-12345 number from a commit message"""
    mdl = None
    match = re.match(r'MDL(-|_)([0-9]+)', message, re.I)
    if match:
        mdl = 'MDL-%s' % (match.group(2))
    return mdl


def get_current_user():
    """Attempt to get the currently logged in user"""
    username = 'root'
    try:
        username = os.getlogin()
    except OSError:
        import getpass
        try:
            username = getpass.getuser()
        except:
            pass
    return username


def launchEditor(filepath=None, suffix='.tmp'):
    """Launchs up an editor

    If filepath is passed, the content of the file is used to populate the editor.

    This returns the path to the saved file.
    """
    editor = resolveEditor()
    if not editor:
        raise Exception('Could not locate the editor')
    with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as tmpfile:
        with open(filepath, 'r') as f:
            tmpfile.write(f.read())
            tmpfile.flush()
        subprocess.call([editor, tmpfile.name])
    return tmpfile.name


def getText(suffix='.md', initialText=None):
    """Gets text from the user using an Editor

    This is a shortcut to using launchEditor as it returns text rather
    than the file in which the text entered is stored.

    When the returned value is empty, the user is asked is they want to resume.
    """
    with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as tmpfile:
        if initialText:
            tmpfile.write(initialText)
            tmpfile.flush()
        while True:
            editorFile = launchEditor(suffix=suffix, filepath=tmpfile.name)
            text = None
            with open(editorFile, 'r') as f:
                text = f.read()

            if len(text) <= 0:
                if not yesOrNo('No content detected. Would you like to resume editing?'):
                    return ''
            else:
                return text

def md5file(filepath):
    """Return the md5 sum of a file
    This is terribly memory inefficient!"""
    return hashlib.md5(open(filepath).read()).hexdigest()


def mkdir(path, perms=0755):
    """Creates a directory ignoring the OS umask"""
    oldumask = os.umask(0000)
    os.mkdir(path, perms)
    os.umask(oldumask)


def parseBranch(branch):
    pattern = re.compile(C.get('wording.branchRegex'), flags=re.I)
    result = pattern.search(branch)
    if not result:
        return False

    parsed = {
        'issue': result.group(pattern.groupindex['issue']),
        'version': result.group(pattern.groupindex['version'])
    }
    try:
        parsed['suffix'] = result.group(pattern.groupindex['suffix'])
    except:
        parsed['suffix'] = None
    return parsed


def process(cmd, cwd=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE):
    if type(cmd) != list:
        cmd = shlex.split(str(cmd))
    logging.debug(' '.join(cmd))
    try:
        proc = subprocess.Popen(cmd, cwd=cwd, stdout=stdout, stderr=stderr)
        (out, err) = proc.communicate()
    except KeyboardInterrupt as e:
        proc.kill()
        raise e
    return (proc.returncode, out, err)


def resolveEditor():
    """Try to resolve the editor that the user would want to use.
       This does actually checks if it is executable"""
    editor = C.get('editor')
    if not editor:
        editor = os.environ.get('EDITOR')
    if not editor:
        editor = os.environ.get('VISUAL')
    if not editor and os.path.isfile('/usr/bin/editor'):
        editor = '/usr/bin/editor'
    return editor


def downloadProcessHook(count, size, total):
    """Hook to report the downloading a file using urllib.urlretrieve"""
    if count <= 0:
        return
    downloaded = int((count * size) / (1024))
    total = int(total / (1024)) if total != 0 else '?'
    if downloaded > total:
        downloaded = total
    sys.stderr.write("\r  %sKB / %sKB" % (downloaded, total))
    sys.stderr.flush()


def stableBranch(version):
    if version == 'master':
        return 'master'
    return 'MOODLE_%d_STABLE' % int(version)


class ProcessInThread(threading.Thread):
    """Executes a process in a separate thread"""

    cmd = None
    cwd = None
    stdout = None
    stderr = None
    _kill = False
    _pid = None

    def __init__(self, cmd, cwd=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE):
        threading.Thread.__init__(self)
        if type(cmd) != 'list':
            cmd = shlex.split(str(cmd))
        self.cmd = cmd
        self.cwd = cwd
        self.stdout = stdout
        self.stderr = stderr

    def kill(self):
        os.kill(self._pid, signal.SIGKILL)

    def run(self):
        logging.debug(' '.join(self.cmd))
        proc = subprocess.Popen(self.cmd, cwd=self.cwd, stdout=self.stdout, stderr=self.stderr)
        self._pid = proc.pid
        while True:
            if proc.poll():
                break

            # Reading the output seems to prevent the process to hang.
            if self.stdout == subprocess.PIPE:
                proc.stdout.read(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

from ..plugins import *
from ..command import Command


class PluginCommand(Command):

    _arguments = [
        (
            ['action'],
            {
                'metavar': 'action',
                'help': 'the action to perform',
                'sub-commands':
                    {
                        'download': (
                            {
                                'help': 'download a plugin in the instance'
                            },
                            [
                                (
                                    ['pluginname'],
                                    {
                                        'type': str,
                                        'metavar': 'pluginname',
                                        'default': None,
                                        'help': 'frankenstyle name of the plugin'
                                    }
                                ),
                                (
                                    ['name'],
                                    {
                                        'default': None,
                                        'help': 'name of the instance to work on',
                                        'metavar': 'name',
                                        'nargs': '?'
                                    }
                                ),
                                (
                                    ['-s', '--strict'],
                                    {
                                        'action': 'store_true',
                                        'help': 'prevent the download of a parent version if a file is not found for this branch',
                                    }
                                ),
                                (
                                    ['-f', '--force'],
                                    {
                                        'action': 'store_true',
                                        'help': 'overrides the plugin directory if it already exists'
                                    }
                                ),
                                (
                                    ['-c', '--no-cache'],
                                    {
                                        'action': 'store_true',
                                        'dest': 'nocache',
                                        'help': 'ignore the cached files'
                                    }
                                )
                            ]
                        ),
                        'install': (
                            {
                                'help': 'download and install a plugin'
                            },
                            [
                                (
                                    ['pluginname'],
                                    {
                                        'type': str,
                                        'metavar': 'pluginname',
                                        'default': None,
                                        'help': 'frankenstyle name of the plugin'
                                    }
                                ),
                                (
                                    ['name'],
                                    {
                                        'default': None,
                                        'help': 'name of the instance to work on',
                                        'metavar': 'name',
                                        'nargs': '?'
                                    }
                                ),
                                (
                                    ['-s', '--strict'],
                                    {
                                        'action': 'store_true',
                                        'help': 'prevent the download of a parent version if a file is not found for this branch',
                                    }
                                ),
                                (
                                    ['-f', '--force'],
                                    {
                                        'action': 'store_true',
                                        'help': 'overrides the plugin directory if it already exists'
                                    }
                                ),
                                (
                                    ['-c', '--no-cache'],
                                    {
                                        'action': 'store_true',
                                        'dest': 'nocache',
                                        'help': 'ignore the cached files'
                                    }
                                )
                            ]
                        )
                    }
            }
        )
    ]
    _description = 'Manage your plugins'

    def run(self, args):

        if args.action == 'download':

            M = self.Wp.resolve(args.name)
            if not M:
                raise Exception('This is not a Moodle instance')

            self.download(M, args)

        elif args.action == 'install':

            M = self.Wp.resolve(args.name)
            if not M:
                raise Exception('This is not a Moodle instance')

            if self.download(M, args):
                logging.info('Upgrading Moodle to install the new plugin')
                M.upgrade()

    def download(self, M, args):

        po = PluginObject(args.pluginname)
        if not args.force and PluginManager.hasPlugin(po, M):
            logging.error('The plugin is already present, set --force to overwrite it.')
            return False

        branch = M.get('branch')
        if branch == 'master':
            branch = C.get('masterBranch')
        branch = int(branch)

        fi = po.getZip(branch, fileCache=not args.nocache)
        if not fi and not args.strict:
            fi = po.getZip(branch - 1, fileCache=not args.nocache)

        if not fi:
            logging.warning('Could not find a file for this plugin')
            return False

        PluginManager.extract(fi, po, M, override=args.force)
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
from .. import tools, css, jira
from ..command import Command
from ..tools import yesOrNo


class BackportCommand(Command):

    _description = 'Backports a branch'

    def __init__(self, *args, **kwargs):
        super(BackportCommand, self).__init__(*args, **kwargs)
        self._arguments = [
            (
                ['-b', '--branch'],
                {
                    'help': 'the branch to backport if not the current one. If omitted, guessed from instance name.',
                    'metavar': 'branch'
                }
            ),
            (
                ['-f', '--force-push'],
                {
                    'action': 'store_true',
                    'dest': 'forcepush',
                    'help': 'force the push'
                }
            ),
            (
                ['-i', '--integration'],
                {
                    'action': 'store_true',
                    'help': 'backport to integration instances'
                }
            ),
            (
                ['-p', '--push'],
                {
                    'action': 'store_true',
                    'help': 'push the branch after successful backport'
                }
            ),
            (
                ['--push-to'],
                {
                    'dest': 'pushremote',
                    'help': 'the remote to push the branch to. Default is %s.' % self.C.get('myRemote'),
                    'metavar': 'remote'
                }
            ),
            (
                ['--patch'],
                {
                    'action': 'store_true',
                    'dest': 'patch',
                    'help': 'instead of pushing to a remote, this will upload a patch file to the tracker. Security issues use this by default if --push is set. This option discards most other flags.',
                }
            ),
            (
                ['-t', '--update-tracker'],
                {
                    'const': True,
                    'dest': 'updatetracker',
                    'help': 'to use with --push, also add the diff information to the tracker issue',
                    'metavar': 'gitref',
                    'nargs': '?'
                }
            ),
            (
                ['-v', '--versions'],
                {
                    'choices': [str(x) for x in range(13, int(self.C.get('masterBranch')))] + ['master'],
                    'help': 'versions to backport to',
                    'metavar': 'version',
                    'nargs': '+',
                    'required': True
                }
            ),
            (
                ['name'],
                {
                    'default': None,
                    'help': 'name of the instance to backport from. Can be omitted if branch is specified.',
                    'metavar': 'name',
                    'nargs': '?'
                }
            )
        ]

    def run(self, args):
        M = None
        branch = args.branch
        versions = args.versions
        integration = args.integration

        # If we don't have a branch, we need an instance
        M = self.Wp.resolve(args.name)
        if not M and not branch:
            raise Exception('This is not a Moodle instance')

        # Getting issue number
        if M and not branch:
            branch = M.currentBranch()

        # Parsing the branch
        parsedbranch = tools.parseBranch(branch)
        if not parsedbranch:
            raise Exception('Could not extract issue number from %s' % branch)
        issue = parsedbranch['issue']
        suffix = parsedbranch['suffix']
        version = parsedbranch['version']

        if args.push and not args.patch:
            mdlIssue = 'MDL-%s' % (issue)
            J = jira.Jira()
            args.patch = J.isSecurityIssue(mdlIssue)
            if args.patch:
                args.push = False
                logging.info('%s appears to be a security issue, switching to patch mode...' % (mdlIssue))

        # Original track
        originaltrack = tools.stableBranch(version)

        # Integration?
        if M:
            integration = M.isIntegration()

        def stashPop(stash):
            """Small helper to pop the stash has we have to do it in some different places"""
            if not stash[1].startswith('No local changes'):
                pop = M2.git().stash(command='pop')
                if pop[0] != 0:
                    logging.error('An error ocured while unstashing your changes')
                else:
                    logging.info('Popped the stash')

        # Begin backport
        for v in versions:

            # Gets the instance to cherry-pick to
            name = self.Wp.generateInstanceName(v, integration=integration)
            if not self.Wp.isMoodle(name):
                logging.warning('Could not find instance %s for version %s' % (name, v))
                continue
            M2 = self.Wp.get(name)

            logging.info("Preparing cherry-pick of %s/%s in %s" % (M.get('identifier'), branch, name))

            # Get hash list
            cherry = '%s/%s..%s' % (self.C.get('upstreamRemote'), originaltrack, branch)
            hashes = M.git().hashes(cherry)
            hashes.reverse()

            # Stash
            stash = M2.git().stash(untracked=False)
            if stash[0] != 0:
                logging.error('Error while trying to stash your changes. Skipping %s.' % M2.get('identifier'))
                logging.debug(stash[2])
                continue
            elif not stash[1].startswith('No local changes'):
                logging.info('Stashed your local changes')

            # Fetch the remote to get reference to the branch to backport
            logging.info("Fetching remote %s..." % (M.get('path')))
            M2.git().fetch(M.get('path'), branch)

            # Creates a new branch if necessary
            newbranch = M2.generateBranchName(issue, suffix=suffix)
            track = '%s/%s' % (self.C.get('upstreamRemote'), M2.get('stablebranch'))
            if not M2.git().hasBranch(newbranch):
                logging.info('Creating branch %s' % newbranch)
                if not M2.git().createBranch(newbranch, track=track):
                    logging.error('Could not create branch %s tracking %s in %s' % (newbranch, track, name))
                    stashPop(stash)
                    continue
                M2.git().checkout(newbranch)
            else:
                M2.git().checkout(newbranch)
                logging.info('Hard reset %s to %s' % (newbranch, track))
                M2.git().reset(to=track, hard=True)

            # Picking the diff upstream/MOODLE_23_STABLE..github/MDL-12345-master
            logging.info('Cherry-picking %s' % (cherry))
            result = M2.git().pick(hashes)
            if result[0] != 0:

                # Try to resolve the conflicts if any.
                resolveConflicts = True
                conflictsResolved = False
                while resolveConflicts:

                    # Check the list of possible conflicting files.
                    conflictingFiles = M2.git().conflictingFiles()
                    if conflictingFiles and len(conflictingFiles) == 1 and 'theme/bootstrapbase/style/moodle.css' in conflictingFiles:
                        logging.info('Conflicts found in bootstrapbase moodle CSS, trying to auto resolve...')
                        cssCompiler = css.Css(M2)
                        if cssCompiler.compile(theme='bootstrapbase', sheets=['moodle']):
                            M2.git().add('theme/bootstrapbase/style/moodle.css')
                            # We need to commit manually to prevent the editor to open.
                            M2.git().commit(filepath='.git/MERGE_MSG')
                            result = M2.git().pick(continu=True)
                            if result[0] == 0:
                                resolveConflicts = False
                                conflictsResolved = True
                    else:
                        resolveConflicts = False

                # We still have a dirty repository.
                if not conflictsResolved:
                    logging.error('Error while cherry-picking %s in %s.' % (cherry, name))
                    logging.debug(result[2])
                    if yesOrNo('The cherry-pick might still be in progress, would you like to abort it?'):
                        result = M2.git().pick(abort=True)
                        if result[0] > 0 and result[0] != 128:
                            logging.error('Could not abort the cherry-pick!')
                        else:
                            stashPop(stash)
                    logging.info('')
                    continue

            # Pushing branch
            if args.push:
                pushremote = args.pushremote
                if pushremote == None:
                    pushremote = self.C.get('myRemote')
                logging.info('Pushing %s to %s' % (newbranch, pushremote))
                result = M2.git().push(remote=pushremote, branch=newbranch, force=args.forcepush)
                if result[0] != 0:
                    logging.warning('Error while pushing to remote %s' % (pushremote))
                    logging.debug(result[2])
                    stashPop(stash)
                    continue

                # Update the tracker
                if args.updatetracker != None:
                    ref = None if args.updatetracker == True else args.updatetracker
                    M2.updateTrackerGitInfo(branch=newbranch, ref=ref)

            elif args.patch:
                if not M2.pushPatch(newbranch):
                    continue

            stashPop(stash)

            logging.info('Instance %s successfully patched!' % name)
            logging.info('')

        logging.info('Done.')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2014 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import os
import jira
import logging


class Fetch(object):
    """Holds the logic and processing to fetch a remote and following actions into an instance"""

    _M = None
    _ref = None
    _repo = None

    _canCreateBranch = True
    _hasstashed = False

    def __init__(self, M, repo=None, ref=None):
        self._M = M
        self._repo = repo
        self._ref = ref

    def checkout(self):
        """Fetch and checkout the fetched branch"""
        self.fetch()
        logging.info('Checking out branch as FETCH_HEAD')
        if not self.M.git().checkout('FETCH_HEAD'):
            raise FetchException('Could not checkout FETCH_HEAD')

    def fetch(self):
        """Perform the fetch"""
        if not self.repo:
            raise FetchException('The repository to fetch from is unknown')
        elif not self.ref:
            raise FetchException('The ref to fetch is unknown')

        git = self.M.git()
        logging.info('Fetching %s from %s' % (self.ref, self.repo))
        result = git.fetch(remote=self.repo, ref=self.ref)
        if not result:
            raise FetchException('Error while fetching %s from %s' % (self.ref, self.repo))

    def _merge(self):
        """Protected method to merge FETCH_HEAD into the current branch"""
        logging.info('Merging into current branch')
        if not self.M.git().merge('FETCH_HEAD'):
            raise FetchException('Merge failed, resolve the conflicts and commit')

    def pull(self, into=None, track=None):
        """Fetch and merge the fetched branch into a branch passed as param"""
        self._stash()

        try:
            self.fetch()
            git = self.M.git()

            if into:
                logging.info('Switching to branch %s' % (into))

                if not git.hasBranch(into):
                    if self.canCreateBranch:
                        if not git.createBranch(into, track=track):
                            raise FetchException('Could not create the branch %s' % (into))
                    else:
                        raise FetchException('Branch %s does not exist and create branch is forbidden' % (into))

                if not git.checkout(into):
                    raise FetchException('Could not checkout branch %s' % (into))

            self._merge()

        except FetchException as e:
            if self._hasstashed:
                logging.warning('An error occured. Some files may have been left in your stash.')
            raise e

        self._unstash()

    def setRef(self, ref):
        """Set the reference to fetch"""
        self._ref = ref

    def setRepo(self, repo):
        """Set the repository to fetch from"""
        self._repo = repo

    def _stash(self):
        """Protected method to stash"""
        stash = self.M.git().stash(untracked=True)
        if stash[0] != 0:
            raise FetchException('Error while trying to stash your changes')
        elif not stash[1].startswith('No local changes'):
            logging.info('Stashed your local changes')
            self._hasstashed = True

    def _unstash(self):
        """Protected method to unstash"""
        if self._hasstashed:
            pop = self.M.git().stash(command='pop')
            if pop[0] != 0:
                logging.error('An error ocured while unstashing your changes')
            else:
                logging.info('Popped the stash')
        self._hasstashed = False

    @property
    def canCreateBranch(self):
        return self._canCreateBranch

    @property
    def into(self):
        return self._into

    @property
    def M(self):
        return self._M

    @property
    def ref(self):
        return self._ref

    @property
    def repo(self):
        return self._repo


class FetchTracker(Fetch):
    """Pretty dodgy implementation of Fetch to work with the tracker.

        If a list of patches is set, we override the git methods to fetch from a remote
        to use the patches instead. I am not super convinced by this design, but at
        least the logic to fetch/pull/merge is more or less self contained.
    """

    _J = None
    _cache = None
    _patches = None

    def __init__(self, *args, **kwargs):
        super(FetchTracker, self).__init__(*args, **kwargs)
        self._J = jira.Jira()
        self._cache = {}

    def checkout(self):
        if not self.patches:
            return super(FetchTracker, self).checkout()

        self.fetch()

    def fetch(self):
        if not self.patches:
            return super(FetchTracker, self).fetch()

        for patch in self.patches:
            j = 0
            dest = None
            while True:
                downloadedTo = patch.get('filename') + (('.' + str(j)) if j > 0 else '')
                dest = os.path.join(self.M.get('path'), downloadedTo)
                j += 1
                if not os.path.isfile(dest):
                    patch['downloadedTo'] = downloadedTo
                    break
            
            logging.info('Downloading patch as %s' % (patch.get('downloadedTo')))
            if not dest or not self.J.download(patch.get('url'), dest):
                raise FetchTrackerException('Failed to download the patch to %s' % (dest))


    def _merge(self):
        if not self.patches:
            return super(FetchTracker, self)._merge()

        patchList = [patch.get('downloadedTo') for patch in self.patches]
        git = self.M.git()
        if not git.apply(patchList):
            raise FetchTrackerException('Could not apply the patch(es), please apply manually')
        else:
            for f in patchList:
                os.remove(f)
        logging.info('Patches applied successfully')

    def getPullInfo(self, mdl):
        """Return the pull information

            This implements its own local cache because we could potentially
            call it multiple times during the same request. This is bad though.
        """
        if not self._cache.has_key(mdl):
            issueInfo = self.J.getPullInfo(mdl)
            self._cache[mdl] = issueInfo
        return self._cache[mdl]

    def setFromTracker(self, mdl, branch):
        """Sets the repo and ref according to the tracker information"""
        issueInfo = self.getPullInfo(mdl)

        repo = issueInfo.get('repo', None)
        if not repo:
            raise FetchTrackerRepoException('Missing information about the repository to pull from on %s' % (mdl))

        ref = issueInfo.get('branches').get(str(branch), None)
        if not ref:
            raise FetchTrackerBranchException('Could not find branch info on %s' % (str(branch), mdl))

        self.setRepo(repo)
        self.setRef(ref.get('branch'))

    def usePatches(self, patches):
        """List of patches (returned by jira.Jira.getAttachments) to work with instead of the standard repo and ref"""
        self._patches = patches

    @property
    def J(self):
        return self._J

    @property
    def patches(self):
        return self._patches


class FetchException(Exception):
    pass


class FetchTrackerException(FetchException):
    pass


class FetchTrackerBranchException(FetchTrackerException):
    pass


class FetchTrackerRepoException(FetchTrackerException):
    pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Moodle Development Kit

Copyright (c) 2013 Frédéric Massart - FMCorz.net

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

http://github.com/FMCorz/mdk
"""

import logging
from .. import tools, jira
from ..command import Command
from ..tools import getMDLFromCommitMessage


class PushCommand(Command):

    _description = 'Push a branch to a remote'

    def __init__(self, *args, **kwargs):
        super(PushCommand, self).__init__(*args, **kwargs)
        self._arguments = [
            (
                ['-b', '--branch'],
                {
                    'metavar': 'branch',
                    'help': 'the branch to push. Default is current branch.'
                }
            ),
            (
                ['-f', '--force'],
                {
                    'action': 'store_true',
                    'help': 'force the push (does not apply on the stable branch)'
                }
            ),
            (
                ['-r', '--remote'],
                {
                    'help': 'remote to push to. Default is your remote.',
                    'default': self.C.get('myRemote'),
                    'metavar': 'remote'
                }
            ),
            (
                ['-p', '--patch'],
                {
                    'action': 'store_true',
                    'help': 'instead of pushing to a remote, this will upload a patch file to the tracker. Security issues use this by default. This option discards most other flags.'
                }
            ),
            (
                ['-t', '--update-tracker'],
                {
                    'const': True,
                    'dest': 'updatetracker',
                    'help': 'also add the diff information to the tracker issue. If gitref is passed, it is used as a starting point for the diff URL.',
                    'metavar': 'gitref',
                    'nargs': '?'
                }
            ),
            (
                ['-s', '--include-stable'],
                {
                    'action': 'store_true',
                    'dest': 'includestable',
                    'help': 'also push the stable branch (MOODLE_xx_STABLE, master)'
                }
            ),
            (
                ['-k', '--force-stable'],
                {
                    'action': 'store_true',
                    'dest': 'forcestable',
                    'help': 'force the push on the stable branch'
                }
            ),
            (
                ['name'],
                {
                    'default': None,
                    'help': 'name of the instance to work on',
                    'metavar': 'name',
                    'nargs': '?'
                }
            )
        ]

    def run(self, args):

        M = self.Wp.resolve(args.name)
        if not M:
            raise Exception('This is not a Moodle instance')

        # Setting remote
        remote = args.remote

        # Setting branch
        if args.branch == None:
            branch = M.currentBranch()
            if branch == 'HEAD':
                raise Exception('Cannot push HEAD branch')
        else:
            branch = args.branch

        # Extra test to see if the commit message is correct. This prevents easy typos in branch or commit messages.
        parsedbranch = tools.parseBranch(branch)
        if parsedbranch or branch != M.get('stablebranch'):
            message = M.git().messages(count=1)[0]

            mdl = getMDLFromCommitMessage(message)

            if parsedbranch:
                branchmdl = 'MDL-%s' % (parsedbranch['issue'])
            else:
                branchmdl = branch

            if not mdl or mdl != branchmdl:
                if not mdl:
                    print 'The MDL number could not be found in the commit message.'
                    print 'Commit: %s' % (message)

                elif mdl != branchmdl:
                    print 'The MDL number in the last commit does not match the branch being pushed to.'
                    print 'Branch: \'%s\' vs. commit: \'%s\'' % (branchmdl, mdl)

                answer = tools.question('Are you sure you want to continue?', default='n')
                if answer.lower()[0] != 'y':
                    print 'Exiting...'
                    return

        J = jira.Jira()

        # If the mode is not set to patch yet, and we can identify the MDL number.
        if not args.patch and parsedbranch:
            mdlIssue = 'MDL-%s' % (parsedbranch['issue'])
            try:
                args.patch = J.isSecurityIssue(mdlIssue)
                if args.patch:
                    logging.info('%s appears to be a security issue, switching to patch mode...', mdlIssue)
            except jira.JiraIssueNotFoundException:
                # The issue was not found, do not perform
                logging.warn('Could not check if %s is a security issue', mdlIssue)

        if args.patch:
            if not M.pushPatch(branch):
                return

        else:
            # Pushing current branch
            logging.info('Pushing branch %s to remote %s...' % (branch, remote))
            result = M.git().push(remote, branch, force=args.force)
            if result[0] != 0:
                raise Exception(result[2])

            # Update the tracker
            if args.updatetracker != None:
                ref = None if args.updatetracker == True else args.updatetracker
                M.updateTrackerGitInfo(branch=branch, ref=ref)

            # Pushing stable branch
            if args.includestable:
                branch = M.get('stablebranch')
                logging.info('Pushing branch %s to remote %s...' % (branch, remote))
                result = M.git().push(remote, branch, force=args.forcestable)
                if result[0] != 0:
                    raise Exception(result[2])

        logging.info('Done.')
